"""
Apache Software License 2.0

Copyright (c) 2020, 8x8, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

from dynaconf.loaders.vault_loader import list_envs
"""
import pkg_resources
from dynaconf import settings
from loguru import logger


class Conf:
    """Manage application settings"""
    _version: str
    log_level = "INFO"
    _banner = "\n______ __\n" \
              "___  //_/________________  __\n" \
              "__  ,<  __  __ \  __ \_  |/_/\n" \
              "_  /| | _  / / / /_/ /_>  <\n" \
              "/_/ |_| /_/ /_/\____//_/|_|\n"  # noqa: W605

    def __init__(self, loglevel=None) -> None:
        """Constructor for Settings"""
        if hasattr(settings, 'LOG_LEVEL'):
            self.set_loglevel(settings.LOG_LEVEL)
        if loglevel:
            self.set_loglevel(loglevel)
        self._version = pkg_resources.get_distribution("knox").version
        self._settings = settings
        logger.debug(f'{self._banner} 🏰 Knox version {self._version}\n')
        logger.trace(self._settings.dynaconf_banner)
        logger.debug("Settings managed with Dynaconf, learn more at http://github.com/rochacbruno/dynaconf")
        logger.debug(f'dynaconf loaded? {self._settings.configured}')

    @classmethod
    def log_filter(cls, record) -> bool:
        levelno = logger.level(cls.log_level).no
        return record["level"].no >= levelno

    @classmethod
    def set_loglevel(cls, level: str) -> None:
        cls.log_level = level

    @property
    def version(self) -> str:
        return self._version

    @property
    def settings(self) -> settings:
        return self._settings



"""
Apache Software License 2.0

Copyright (c) 2020, 8x8, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""
from .store_object import StoreObject


class StoreEngine:
    """The abstract persistence strategy for storing the certificates """
    settings: object

    def __init__(self) -> None:
        """Constructor for StoreEngine"""

    def open(self) -> bool:
        """Initialize access to the persistence"""
        pass

    def close(self) -> bool:
        """Close access to the persistence"""
        pass

    def initialize(self) -> bool:
        """Ensure the store is configured properly"""
        pass

    def read(self, path: str, name: str, type=None) -> StoreObject:
        """Read from the store"""
        pass

    def write(self, obj: StoreObject) -> bool:
        """Write to the store"""
        pass

    def delete(self, path: str, name: str) -> bool:
        """Delete from the store"""
        pass



"""
Apache Software License 2.0

Copyright (c) 2020, 8x8, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""
import ast
import enum
import json
from binascii import hexlify

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import dsa
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import Encoding
from dynaconf import LazySettings
from jinja2 import Environment
from jinja2 import FileSystemLoader
from loguru import logger

from ..backend import StoreObject
from .cert_engine import CertDnsEngine


class Cert(StoreObject):
    """Object representation of a TLS certificate"""
    _body: str  #: String representation of private, chain and public portions of certificate as a map/json
    _info: str  #: Certificate details
    _data: {}  #: Combined body and info map
    _file: object
    _x509: x509
    _common_name: str
    _jinja: Environment

    class CertTypes(enum.Enum):
        PEM = 1
        DER = 2
        PFX = 3

    PEM = CertTypes.PEM
    DER = CertTypes.DER
    PFX = CertTypes.PFX

    def __init__(self, settings: LazySettings, common_name=None) -> None:
        """Constructor for Cert"""
        self._settings = settings
        self._common_name = self.valid_name(common_name)
        self._body = ""
        self._info = ""
        super().__init__(name=self._common_name, path=self.store_path(), body=self._body, info=self._info)
        self._jinja = Environment(loader=FileSystemLoader('templates'))
        self._tmpl_body = self._jinja.get_template('body_template.js')
        self._tmpl_info = self._jinja.get_template('info_template.js')
        self._tmpl_data = self._jinja.get_template('data_template.js')

    def load_x509(self, path: str) -> None:
        """Given path to PEM x509 read in certificate

            :param path: File path to x509 PEM file
            :type path: str
        """
        with open(path, mode='r+', encoding='utf-8') as fp:
            self._file = fp.read()
            self._x509 = x509.load_pem_x509_certificate(bytes(self._file, 'utf-8'), default_backend())

        """Generate data structures using custom Jinja2 templates"""
        self._info = ast.literal_eval(self._tmpl_info.render(cert=self))
        self._body = ast.literal_eval(self._tmpl_body.render(cert=self))
        self._data = ast.literal_eval(self._tmpl_data.render(cert=self))

        """Ensure raw file contents in public key, Jinja2 fails to parse if there are CR LF"""
        self._body['cert_body']['public'] = self._file
        self._data['cert_body']['public'] = self._file

        """Match the objects common name to the true common name from the certificate and
        swap out '*' astrix for the keyword wildcard
        """
        self._common_name = self.valid_name(self._data['cert_info']['subject']['commonName'])
        self.name = self.valid_name(self._common_name)

        """Ensure path is the inverse of the true cert common name"""
        self.path = self.store_path()

        self.type = Cert.PEM.name

    def load(self, pub: str, key: str, certtype: enum.Enum = PEM, chain: str = None) -> None:
        """Read in components of a certificate, given filename paths for each

            :param pub: File name of public portion of key
            :type pub: str
            :param key: File name of private portion of key
            :type key: str
            :param chain: File name of intermediate certificates. Optional as they could be in pub
            :type chain: str
            :param certtype: Enum of certificate types [PEM=1, DER=2]
            :type certtype: Enum
        """
        if certtype == Cert.PEM.name:
            self.load_x509(pub)

        with open(key, mode="r") as key_fp:
            self._body['cert_body']['private'] = key_fp.read()

        if chain:
            with open(chain, mode="r") as chain_fp:
                self._body['cert_body']['chain'] = chain_fp.read()

        self._data['cert_body'] = self._body['cert_body']

    @classmethod
    def valid_name(cls, value: str) -> str:
        """Some engines might have problems with astrix, as they are used for glob searching and or RBAC.
        Replace it with the key word 'wildcard'. This does not affect the actual certificate."""
        return value.replace('*', 'wildcard')

    def subject(self) -> str:
        """Return the certificate subject details"""
        return json.dumps({attr.oid._name: attr.value for attr in self._x509.subject}, indent=8)

    def issuer(self) -> str:
        """Return the certificate issuer details"""
        return json.dumps({attr.oid._name: attr.value for attr in self._x509.issuer}, indent=8)

    def validity(self) -> str:
        """Return the certificates dates of validity"""
        cert = self._x509
        return json.dumps({
            'not_valid_before': f'{cert.not_valid_before}',
            'not_valid_after': f'{cert.not_valid_after}',
        }, indent=8)

    def key_details(self) -> str:
        """Return characteristics of key used to generate the certificate"""
        cert = self._x509
        public_key = self._x509.public_key()
        key_info = {'size': public_key.key_size}
        if isinstance(public_key, rsa.RSAPublicKey):
            key_type = 'RSA'
        elif isinstance(public_key, dsa.DSAPublicKey):
            key_type = 'DSA'
        elif isinstance(public_key, ec.EllipticCurvePublicKey):
            key_type = 'ECC'
            key_info['curve'] = public_key.curve.name
        else:
            raise ValueError('Invalid key type.')
        key_info['type'] = key_type
        return json.dumps({
            'version': cert.version.name,
            'fingerprint_sha256': hexlify(cert.fingerprint(hashes.SHA256())).decode(),
            'serial_number': f'{cert.serial_number}',
            'key': key_info
        }, indent=8)

    @staticmethod
    def to_store_path(common_name: str) -> str:
        """Generate a backend store path based on the certificates common name
        www.example.com becomes /com/example/www

            :return: str
        """
        domainsplit = common_name.split('.')
        return "/"+"/".join(reversed(domainsplit))

    def store_path(self) -> str:
        return self.to_store_path(self._common_name)

    def __str__(self) -> str:
        return json.dumps(self._data, indent=4)

    def private(self) -> str:
        """Unless its a dict, its not loaded yet"""
        if isinstance(self._body, dict):
            return json.dumps(self._body['cert_body']['private']).replace('\n', '')
        else:
            return ""

    def chain(self) -> str:
        """Unless its a dict, its not loaded yet"""
        if isinstance(self._body, dict):
            return json.dumps(self._body['cert_body']['chain']).replace('\n', '')
        else:
            return ""

    def public(self) -> str:
        """Convenience method for Jinja2 templates. Jinja2 does not process the string if it has carriage returns."""
        return self._x509.public_bytes(Encoding.PEM).decode('utf-8').replace('\n', '')

    def info(self) -> str:
        return json.dumps(self._info['cert_info'], indent=4)

    def body(self) -> str:
        return json.dumps(self._body['cert_body'], indent=4)

    @property
    def data(self) -> str:
        """Content to persist, typically JSON"""
        return self._data

    def generate(self) -> None:
        """ Generate certificate for a given common name"""
        try:
            cde = CertDnsEngine(self._settings)
            certfile, chainfile, privkey = cde.call_provider(self._common_name)
            self.load(pub=certfile, key=privkey, chain=chainfile)
        except Exception:
            logger.error(f'Failed to generate certificate {self._common_name}')




#from click.testing import CliRunner

#from knox import cli


def test_main():
    pass
#    runner = CliRunner()
#    result = runner.invoke(main, [])
#
#    assert result.output == '()\n'
#    assert result.exit_code == 0



"""
Apache Software License 2.0

Copyright (c) 2020, 8x8, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""
from . import backend as backend
from . import config as config


class Knox:
    """Composite class for Knox package"""
    _conf: config.Conf
    _store: backend.Store
    _cli: dict

    def __init__(self, loglevel=None) -> None:
        """Constructor for Knox """

        self._conf = config.Conf(loglevel)
        self._store = backend.Store(self._conf.settings)

    @property
    def settings(self) -> config.Conf.settings:
        """Access to the Dynaconf settings object"""
        return self._conf.settings

    @property
    def conf(self) -> config.Conf:
        """Access to the knox Conf object"""
        return self._conf

    @property
    def store(self) -> backend.Store:
        """Access to the instantiated store engine"""
        return self._store



"""
Apache Software License 2.0

Copyright (c) 2020, 8x8, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -m knox` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``knox.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``knox.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
import sys

import click
import pkg_resources
from loguru import logger

from .certificate import Cert  # noqa: F401
from .config import Conf
from .knox import Knox


@click.group()
@click.option('--debug/--no-debug', default=False)
@click.version_option(version=pkg_resources.get_distribution('knox').version)
@click.pass_context
@logger.catch()
def cli(ctx, debug):
    """Utilities for managing and storing TLS certificates using backing store (Vault)."""
    # ensure that ctx.obj exists and is a dict (in case `cli()` is called
    # by means other than the `if` block below)
    ctx.ensure_object(dict)
    ctx.obj['DEBUG'] = debug
    logger.remove()
    if debug:
        ctx.obj['LOG_LEVEL'] = 'DEBUG'
        logger.add(sys.stdout, format="{time} {level: >9} {level.icon} {message}", filter=Conf.log_filter, level="DEBUG", colorize=True)
        logger.info(f' Log level set to {ctx.obj["LOG_LEVEL"]}')
    else:
        ctx.obj['LOG_LEVEL'] = 'WARNING'
        logger.add(sys.stderr, format="{time} {level: >9} {level.icon} {message}", filter=Conf.log_filter, level="WARNING", colorize=True)


@cli.group(no_args_is_help=True)
@click.option("--type", "-t", type=click.Choice(['PEM', 'DER', 'PFX'], case_sensitive=True), default='PEM', show_default=True)
@click.option("--pub", help="Public key file")
@click.option("--chain", help="Intermediate chain")
@click.option("--key", help="Private key file")
@click.pass_context
@logger.catch()
def cert(ctx, type, pub, key, chain=None):
    """Certificate utilities.

    NAME is the common name for the certificate. i.e. www.example.com
    """
    ctx.obj['CERT_PUB'] = pub
    ctx.obj['CERT_CHAIN'] = chain
    ctx.obj['CERT_KEY'] = key
    ctx.obj['CERT_TYPE'] = type


@cert.command(no_args_is_help=True)
@click.argument("name")
@click.pass_context
@logger.catch()
def save(ctx, name):
    """Store an existing certificate
    """
    ctx.obj['CERT_NAME'] = name
    pub = ctx.obj['CERT_PUB']
    key = ctx.obj['CERT_KEY']
    chain = ctx.obj['CERT_CHAIN']
    certtype = ctx.obj['CERT_TYPE']

    knox = Knox(ctx.obj['LOG_LEVEL'])
    certificate = Cert(knox.settings, common_name=name)
    certificate.load(pub=pub,
                     key=key,
                     chain=chain,
                     certtype=certtype)
    knox.store.save(certificate)


@cert.command(no_args_is_help=True)
@click.argument("name")
@click.pass_context
@logger.catch()
def get(ctx, name):
    """Retrieve an existing certificate for a given common name from the store
    """
    ctx.obj['CERT_NAME'] = name
    knox = Knox(ctx.obj['LOG_LEVEL'])
    certificate = Cert(knox.settings, common_name=name)
    certificate.type = ctx.obj['CERT_TYPE']
    certificate = knox.store.get(certificate.store_path(), name=name, type=certificate.type)
    with open(certificate.name+"-pub.pem", "w") as pubf:
        pubf.write(certificate.body['public'])
    with open(certificate.name+"-key.pem", "w") as keyf:
        keyf.write(certificate.body['private'])
    with open(certificate.name+"-chain.pem", "w") as chainf:
        chainf.write(certificate.body['chain'])


@cert.command(no_args_is_help=True)
@click.argument("name")
@click.pass_context
@logger.catch()
def gen(ctx, name):
    """Create and store a new certificate for a given common name
    """
    ctx.obj['CERT_NAME'] = name

    knox = Knox(ctx.obj['LOG_LEVEL'])
    certificate = Cert(knox.settings, common_name=name)
    certificate.generate()
    knox.store.save(certificate)


@cli.command(no_args_is_help=True)
@click.option("-f", "--find", help="Find certificate by common name")
@click.argument("name")
@click.pass_context
@logger.catch()
def store(ctx, find, name) -> dict:
    """Store commands. Given a certificate NAME perform the store operation.

    NAME can be similar to a full file path or the certificates common name.
    i.e. www.example.com becomes /com/example/www/www.example.com when stored.

    """
    ctx.obj['STORE_FIND'] = find
    knox = Knox(ctx.obj['LOG_LEVEL'])
    if find:
        certificate = knox.store.find(Cert.to_store_path(name), name=name)  # noqa F841
        # save certificate_public_key.pem
        # save certificate_private_key.pem
        # save certificate_chain.pem

    return ctx


def main():
    cli(prog_name="knox", obj={})


if __name__ == "__main__":
    main()



"""
Apache Software License 2.0

Copyright (c) 2020, 8x8, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import os
import subprocess

from loguru import logger


class CertDnsEngine:
    """Data driven certificate verification DNS providers"""

    def __init__(self, settings):
        """DNS Engine for generating certificates

            :param settings: dynaconf settings
            :type settings: json
        """
        self._settings = settings
        """[TODO 5/30/20] ljohnson use dynaconf's Vault backend to load these credentials"""
        self._domain_metadata = self._settings.as_dict()['KNOX_DNS_METADATA']
        self._certbot_conf = self._settings.as_dict()['KNOX_CERTBOT_CONFIG']

    def get_provider(self, common_name: str) -> str:
        """Find the DNS provider plugin that is authoritative for the common name

            What DNS service is being used as the authoritative server of the
            given domain? Route53? Cloudflare? etc.

            This is required to complete the ACME ownership challenge with certbot.

            Provider credentials and associating domains are configurable via KNOX_DNS_METADATA.

            See config/settings.json

            :param common_name: The common name being used to generate a certificate
            :type common_name: str
            :returns: str
        """
        domain_suffix = '.'.join(common_name.split('.')[1:])

        """Validate if a domain name derived from common name is supported"""
        for provider in self._domain_metadata:
            domains = self._domain_metadata[provider]['domains']
            if domain_suffix in domains:
                logger.info(f' Domain name {domain_suffix} supported by {provider}')
                return provider

        logger.error(f'No provider found supporting {domain_suffix}')

    def validate_provider_credentials(self, provider: str) -> bool:
        """Validate if provider credentials are present

            Every DNS provider have a set of credentials to authenticate.
            This function validates if the required credentials are present.

            See config/settings.json for required credentials

            :param provider: Provider name to verify the credentials
            :type provider: str
            :returns: bool
        """

        credentials_list = self._domain_metadata.get(provider).get('required_credentials')
        try:
            for credential in credentials_list:
                if os.environ[credential]:
                    pass
            return True
        except Exception:
            logger.error(f'Provider credentials for {provider} are not set/valid')
            return False

    def provider_cmd(self, plugin) -> str:
        """Generate DNS provider specific certbot commands

            :param plugin: certbot args for specific dns provider
            :type plugin: str
            :returns str
        """
        command = f'certbot certonly -n --agree-tos --{plugin} -d {self._common_name}'
        return command

    def call_provider(self, name: str) -> tuple:
        """Generate certificate using certbot by calling specific provider args

            Get the provider name and verify their credentials.
            If both are set and True, run the certbot commands with
            specific provider plugin arguments

            :param name:  The common name being used to generate a certificate
            :type name: str
            :returns tuple
        """
        root_dir = self._certbot_conf.get('KNOX_CERTBOT_ROOT_DIR')
        work_dir = f'{root_dir}/var/lib/letsencrypt'
        logs_dir = f'{root_dir}/var/log/letsencrypt'
        conf_dir = f'{root_dir}/etc/letsencrypt'
        email_id = self._certbot_conf.get('KNOX_CERTBOT_ACCOUNT_EMAIL')
        provider_name = self.get_provider(name)
        if provider_name:
            if self.validate_provider_credentials(provider_name):
                logger.info(f' Calling provider {provider_name} for {name}')
                provider_cmd = self.provider_cmd(provider_name)
                certbot_command = f'{provider_cmd} ' \
                                  f'--email {email_id} ' \
                                  f'--work-dir {work_dir} ' \
                                  f'--logs-dir {logs_dir} ' \
                                  f'--config-dir {conf_dir}'
                response = subprocess.run(certbot_command,
                                          shell=True,
                                          stdout=subprocess.PIPE,
                                          stderr=subprocess.STDOUT)
                if response.returncode == 0:
                    certfile_path = f'{conf_dir}/live/{name}/cert.pem'
                    chainfile_path = f'{conf_dir}/live/{name}/chain.pem'
                    privkey_path = f'{conf_dir}/live/{name}/privkey.pem'
                    return certfile_path, chainfile_path, privkey_path
                else:
                    logger.error(f'Certbot return code {response.returncode} none zero')
            else:
                logger.error(f'Credentials not found for {provider_name} provider')
        else:
            logger.error(f'No DNS provider found for {name}')
